"""
This script defines various feature extractors used for feature-based image encoders 
such as VLAD (Vector of Locally Aggregated Descriptors) or Fisher Vectors. It includes 
implementations for handcrafted features (e.g., SIFT, RootSIFT), user-defined 
custom feature extraction functions, and deep convolutional feature extraction 
with optional spatial encoding.
"""

from functools import wraps
from typing import Callable

import cv2
import numpy as np
import torch
from torchvision import transforms
from torchvision.models import vgg16, VGG16_Weights

from .._base_classes import FeatureExtractorBase
from .._config import setup_logging
from .._utils import check_is_image

setup_logging()


def _check_output_shape(func) -> Callable:
    """
    Ensures the feature extractor output is a 2D NumPy array of shape
    (num_vectors, self.output_dim).
    """
    @wraps(func)
    def wrapper(self, *args, **kwargs) -> np.ndarray:
        feat_vecs = func(self, *args, **kwargs)
        if feat_vecs is None:
            print("No feature vectors found. Returning empty array.")
            return np.zeros((0, self.output_dim), dtype=np.float32)

        if not isinstance(feat_vecs, np.ndarray):
            raise ValueError(f"Expected output to be a NumPy array, got {type(feat_vecs)} instead.")

        if feat_vecs.ndim != 2:
            raise ValueError(f"Feature extractor output must be 2D. Got shape {feat_vecs.shape}.")

        if feat_vecs.shape[1] != self.output_dim:
            raise ValueError(f"Expected feat_vecs.shape[1] == {self.output_dim}, "
                             f"but got {feat_vecs.shape[1]}.")

        return feat_vecs

    return wrapper


class SIFT(FeatureExtractorBase):
    """
    Scale-Invariant Feature Transform (SIFT) feature extractor.

    References:
    ===========
    [1] Lowe, D. G. (2004). Distinctive image features from scale-invariant keypoints.
    """
    def __init__(self):
        super().__init__()
        self._output_dim = 128

    @property
    def output_dim(self) -> int:
        return self._output_dim

    @_check_output_shape
    @check_is_image(arg_positions=(1,))
    def __call__(self, image: np.ndarray, /) -> np.ndarray:
        """
        Extracts SIFT features from an image.
        :param image:
        :return:
        """
        sift = cv2.SIFT.create()
        _, descriptors = sift.detectAndCompute(image, None)
        return descriptors


class RootSIFT(FeatureExtractorBase):
    """
    Scale-Invariant Feature Transform with Hellinger kernel (RootSIFT) normalizer.

    References:
    ===========
    [1] Arandjelovic, R., & Zisserman, A. (2012). Three things everyone should know to improve object retrieval.
    """
    def __init__(self):
        super().__init__()
        self._output_dim = 128

    @property
    def output_dim(self) -> int:
        return self._output_dim

    @_check_output_shape
    @check_is_image(arg_positions=(1,))
    def __call__(self, image: np.ndarray,/) -> np.ndarray:
        """
        Extracts RootSIFT features from an image.
        :param image:
        :return:
        """
        sift = cv2.SIFT.create()
        _, descriptors = sift.detectAndCompute(image, None)
        descriptors /= (descriptors.sum(axis=1, keepdims=True) + 1e-7)
        descriptors = np.sqrt(descriptors)
        return descriptors


class Lambda(FeatureExtractorBase):
    """
    Lambda feature extractor that allows passing any user-defined
    function to extract features from images.

    The function must accept a single argument (image as NumPy array),
    and output fixed-size feature vectors from each image.
    """
    def __init__(self, func: Callable, output_dim: int):
        """
        Initializes the Lambda feature extractor.
        :param func:
        :param output_dim:
        """
        super().__init__()
        if not callable(func):
            raise ValueError(f"Argument func must be a callable object, got {type(func)} instead")
        self._output_dim = output_dim
        self.func = func

    @property
    def output_dim(self) -> int:
        return self._output_dim

    @_check_output_shape
    @check_is_image(arg_positions=(1,))
    def __call__(self, image: np.ndarray, /) -> np.ndarray:
        return self.func(image)


class DeepConvFeature(FeatureExtractorBase):
    """
    Extracts convolutional feature maps from a chosen conv layer of a torchvision model.
    It flattens the feature maps into feature descriptors. Optionally appends
    normalized (x, y) coordinates to each spatial location.

    The concepts here were inspired by by the work on `VLAD-DCNN` features for face verification, as
    presented in [1], where VLAD encodings were computed from deep convolutional features and input into
    a metric learning algorithm in order to distinguish between different people.

    :param model: A PyTorch model instance from torchvision.model_files. Default is VGG16. In the paper [1], a VGG-Face model trained on Idmb-Wiki
    dataset was used with VLAD encoding for younger faces verification.
    :param target_submodule: Optional submodule name to hook into. If None, the whole model is used.
    :param layer_index: Which conv layer to hook (int). Use `list_conv_layers(...)`
                       to see the ordering or use -1 for the last conv layer.
    :param spatial_encoding: If True, appends (x/W, y/H) to each descriptor.
    :param device: 'cpu' or 'cuda'. Where to run the model.
    :param transform: Optional torchvision.transforms.Compose. Default includes `to_tensor`, `resize(224, 224)`,
                        and normalization with ImageNet stats.

    References:
    ===========
    [1] Liangliang Wang and Deepu Rajan, "An Image Similarity Descriptor for Classification Tasks," J. Vis. Commun. Image R., vol. 71, pp. 102847, 2020.
    [2] Weixia Zhang, Jia Yan, Wenxuan Shi, Tianpeng Feng, and Dexiang Deng, "Refining Deep Convolutional Features for Improving Fine-Grained Image 
    Recognition," EURASIP Journal on Image and Video Processing, 2017.
    """
    def __init__(
        self,
        model: torch.nn.Module = vgg16(weights=VGG16_Weights.DEFAULT),
        target_submodule: str = None,
        layer_index: int = -1,
        spatial_encoding: bool = True,
        device: str = torch.device("cuda" if torch.cuda.is_available() else "cpu"),
        transform: transforms.Compose = None
    ):
        super().__init__()
        self._model = None
        self.layer_index = layer_index
        self.spatial_encoding = spatial_encoding
        self.device = device
        self.transform = transform
        if self.transform is None:
            self.transform = transforms.Compose([transforms.ToTensor(),
                                            transforms.Resize((224, 224))])

        self.model: torch.nn.Module  = model # Trigger setter
        self._modules: torch.nn.Module = self._get_submodule(target_submodule)
        self._conv_layers = self.list_conv_layers()
        if not self._conv_layers:
            raise ValueError(f"No convolutional layers found in model {self.model._get_name()}.")

        self.buffer = None
        try:
            _, self.selected_layer_name, self.selected_layer_module = self._conv_layers[self.layer_index]
            self._logger.info(f"Selected layer: {self.selected_layer_name}, {self.selected_layer_module}")
        except IndexError:
            info = "" if target_submodule is None else f" in submodule {self._modules._get_name()}"
            raise IndexError(f"Model {self.model._get_name()} has only {len(self._conv_layers)} convolutional layers {info}"
                             f". Got layer_index={self.layer_index}.")
        self._output_dim = self.selected_layer_module.out_channels + 2 if self.spatial_encoding else self.selected_layer_module.out_channels
        self._register_hook()

    @property
    def output_dim(self) -> int:
        return self._output_dim

    @property
    def model(self) -> torch.nn.Module:
        return self._model

    @model.setter
    def model(self, model: torch.nn.Module):
        if not isinstance(model, torch.nn.Module):
            raise TypeError(f"Currently, only torch.nn.Module is supported. Got {type(model)} instead.")
        self._model = model

    def _get_submodule(self, submodule_name: str = None) -> torch.nn.Module:
        """
        Retrieves a submodule from a PyTorch model by name.

        :return: The submodule instance.
        """
        if submodule_name is None:
            return self._model
        if not hasattr(self._model, submodule_name):
            raise AttributeError(f"Model {self.model._get_name()} has no submodule named {submodule_name}.")
        return getattr(self._model, submodule_name)

    def list_conv_layers(self) -> list[tuple[int, str, torch.nn.Module]]:
        """
        Utility function to collect convolutional layers (and sub-modules)
        from the model / chosen submodule.

        :return: List of (layer_index, layer_module) for each convolutional layer.
        """
        conv_layers = []
        idx = 0
        for name, module in self._modules.named_modules():
            if isinstance(module, torch.nn.Conv2d):
                conv_layers.append((idx, name, module))
                idx += 1
        return conv_layers

    def _register_hook(self):
        """
        Registers a forward hook on the selected convolutional layer
        to capture its output (feature map).
        """
        def hook_fn(module, input, output):
            self.buffer = output.detach() # output shape: [batch_size, channels, height, width]
        self.hook = self.selected_layer_module.register_forward_hook(hook_fn)

    @_check_output_shape
    @check_is_image(arg_positions=(1,))
    def __call__(self, image: np.ndarray, /) -> np.ndarray:
        """
        #TODO: first, check if image is tensor and has range [0,1]. If numpy and has range [0,255], normalize and convert to tensor. If numpy and has range [0,1], convert to tensor. Else, raise error.
        Processes a single image through the chosen conv layer and
        returns flattened feature descriptors.

        :param image: Input image as a NumPy array (H x W x C, BGR or RGB).
        :return: N x D NumPy array, where N = (H_conv x W_conv) and
                 D = number_of_channels (+ 2 if spatial coords are appended).
        """
        image = self.transform(image).unsqueeze(0).to(self.device)

        self.model.eval()
        self.model.to(self.device)
        _ = self.model(image)  # we only care about the hook's output
        if self.buffer is None:
            raise RuntimeError("Forward hook did not capture any features.")

        # Convert the captured feature map to NumPy
        feature_map = self.buffer.cpu().numpy()  # shape: (1, C, Hf, Wf)
        feature_map = feature_map[0]  # Remove batch dimension

        C, Hf, Wf = feature_map.shape
        feature_map = feature_map.reshape(C, -1).T  # shape: (Hf*Wf, C)

        if self.spatial_encoding:
            coords = []
            for y in range(Hf):
                for x in range(Wf):
                    coords.append([x / Wf, y / Hf])  # (x/Wf, y/Hf)
            coords = np.array(coords, dtype=np.float32)  # shape: (Hf*Wf, 2)
            # Concatenate
            feature_map = np.hstack([feature_map, coords]) # shape: (Hf*Wf, C+2)

        return feature_map


import os
from typing import Optional
from platformdirs import user_cache_dir
from multiprocessing import Process
import logging

import cv2
import numpy as np
import requests
import scipy
from torch.utils.data import Dataset
from torchvision import transforms
from tqdm import tqdm

from pyvisim._config import *

setup_logging()

__all__ = ['OxfordFlowerDataset']

logger = logging.getLogger('Data_Set')

# Specific to the Oxford Flowers dataset
_DATASET_ROOT = os.path.join(user_cache_dir("pyvisim"), "oxford_flower_dataset")
_IMAGE_DIR = os.path.join(_DATASET_ROOT, "images/jpg")
_IMAGE_LABEL_FILE = os.path.join(_DATASET_ROOT, "labels.mat")
_SETID_FILE = os.path.join(_DATASET_ROOT, "setid.mat")
OXFORD_NUM_IMAGES = 8189
NUM_TEST_IMG = 6149
NUM_TRAIN_IMG = 1020
NUM_VAL_IMG = 1020

def _download_and_process_file(url: str, dest: str, extract_dir: str):
    """
    Downloads a file and processes it (e.g., extraction if it's a zip or tar.gz file).
    """
    _download_file_with_progress(url, dest)

    if dest.endswith(".zip"):
        _extract_zip(dest, os.path.join(extract_dir, os.path.splitext(os.path.basename(dest))[0]))
        os.remove(dest)
    elif dest.endswith(".tgz") or dest.endswith(".tar.gz"):
        _extract_tar(dest, os.path.join(extract_dir, os.path.splitext(os.path.basename(dest))[0]))
        os.remove(dest)

def _download_file_with_progress(url: str, dest: str):
    """
    Download a file with a progress bar.
    """
    logger.info(f"Downloading from {url} to {dest}")
    response = requests.get(url, stream=True)
    total_size = int(response.headers.get('content-length', 0))

    with open(dest, "wb") as f, tqdm(
        total=total_size, unit='B', unit_scale=True, desc=f"Downloading {os.path.basename(dest)}"
    ) as progress_bar:
        for chunk in response.iter_content(chunk_size=8192):
            if chunk:
                f.write(chunk)
                progress_bar.update(len(chunk))

    logger.info(f"Downloaded file to {dest}")

def _extract_zip(zip_file: str, extract_to: str):
    """
    Extract a zip archive.
    """
    import zipfile
    logger.info(f"Extracting {zip_file} to {extract_to}")
    with zipfile.ZipFile(zip_file, 'r') as zf:
        total_files = len(zf.infolist())
        with tqdm(total=total_files, unit='file', desc=f"Extracting {os.path.basename(zip_file)}") as progress_bar:
            for file in zf.infolist():
                zf.extract(file, extract_to)
                progress_bar.update(1)

def _extract_tar(tar_file: str, extract_to: str):
    """
    Extract a tar.gz archive.
    """
    import tarfile
    logger.info(f"Extracting {tar_file} to {extract_to}")
    with tarfile.open(tar_file, "r:gz") as tf:
        members = tf.getmembers()
        total_files = len(members)
        with tqdm(total=total_files, unit='file', desc=f"Extracting {os.path.basename(tar_file)}") as progress_bar:
            for member in members:
                tf.extract(member, path=extract_to)
                progress_bar.update(1)

def _data_downloaded() -> bool:
    """
    Check if the image files, labels, and setid files are downloaded.
    """
    if not os.path.isdir(_DATASET_ROOT):
        return False
    if not os.path.isdir(_IMAGE_DIR):
        return False
    if not os.path.isfile(_IMAGE_LABEL_FILE) or not os.path.isfile(_SETID_FILE):
        return False
    return True

def _check_data_integrity() -> bool:
    """
    Checks if the downloaded data are correct:
      1) labels.mat has exactly OXFORD_NUM_IMAGES labels
      2) setid.mat has correct lengths for tstid, valid, trnid
      3) images/ has exactly OXFORD_NUM_IMAGES images
    Returns True if all checks pass, False otherwise.
    """
    if not os.path.isfile(_IMAGE_LABEL_FILE):
        logger.warning(f"Label file not found at {_IMAGE_LABEL_FILE}.")
        return False
    try:
        mat_data = scipy.io.loadmat(_IMAGE_LABEL_FILE)
        labels = mat_data['labels'].squeeze().tolist()
        if len(labels) != OXFORD_NUM_IMAGES:
            logger.warning(f"Expected {OXFORD_NUM_IMAGES} labels, got {len(labels)}.")
            return False
    except Exception as e:
        logger.warning(f"Error reading labels file: {e}")
        return False

    if not os.path.isfile(_SETID_FILE):
        logger.warning(f"setid.mat not found at {_SETID_FILE}.")
        return False
    try:
        mat_data = scipy.io.loadmat(_SETID_FILE)
        tstid = mat_data['tstid'].squeeze().tolist()   # len 6149
        valid = mat_data['valid'].squeeze().tolist()   # len 1020
        trnid = mat_data['trnid'].squeeze().tolist()   # len 1020
        if len(tstid) != NUM_TEST_IMG or len(valid) != NUM_VAL_IMG or len(trnid) != NUM_TRAIN_IMG:
            logger.warning(f"setid.mat has incorrect lengths. tstid={len(tstid)}, "
                           f"valid={len(valid)}, trnid={len(trnid)}."
                           f"Expected {NUM_TEST_IMG}, {NUM_VAL_IMG}, {NUM_TRAIN_IMG} images respectively.")
            return False
    except Exception as e:
        logger.warning(f"Error reading setid file: {e}")
        return False

    if not os.path.isdir(_IMAGE_DIR):
        logger.warning(f"Image directory not found at {_IMAGE_DIR}.")
        return False
    jpgs = [f for f in os.listdir(_IMAGE_DIR) if f.lower().endswith('.jpg')]
    if len(jpgs) != OXFORD_NUM_IMAGES:
        logger.warning(f"Expected {OXFORD_NUM_IMAGES} .jpg images, got {len(jpgs)}.")
        return False
    return True

def download_oxford_flowers_data(destination: str=_DATASET_ROOT):
    """
    Downloads the 102 flowers dataset and organizes it into the desired structure,
    under `destination/oxford_flower_dataset/`.
    """
    logger.info("Starting download process for Oxford Flowers.")
    os.makedirs(_DATASET_ROOT, exist_ok=True)

    files_to_download = {
        "images": "https://www.robots.ox.ac.uk/~vgg/data/flowers/102/102flowers.tgz",
        "labels": "https://www.robots.ox.ac.uk/~vgg/data/flowers/102/imagelabels.mat",
        "setid": "https://www.robots.ox.ac.uk/~vgg/data/flowers/102/setid.mat"
    }

    processes = []
    for name, url in files_to_download.items():
        file_path = os.path.join(_DATASET_ROOT, f"{name}{os.path.splitext(url)[-1]}")
        p = Process(target=_download_and_process_file, args=(url, file_path, _DATASET_ROOT))
        processes.append(p)
        p.start()

    for p in processes:
        p.join()

    logger.info("Oxford Flowers dataset downloaded and processed successfully.")


class OxfordFlowerDataset(Dataset):
    """
    Oxford Flower Dataset. It can be found at: https://www.robots.ox.ac.uk/~vgg/data/flowers/102/index.html.

    In the original dataset, number of train images ('trnid') is 1020,
    number of validation images ('valid') is 1020, and number of test images ('tstid') is 6149. Since
    it makes more sense to have more images for training for this project, the train and test
    splits have been swapped.

    :param transform: Transformations to apply to the images.
    :param purpose: Purpose of the dataset ('train', 'test', 'validation'). You
    can also pass a list such as ['train', 'validation'] to get a combined dataset.
    """
    def __init__(self,
                 transform: Optional[transforms.Compose] = None,
                 purpose: str | list[str] = 'train') -> None:
        self.transform = transform
        self.purpose = [purpose] if isinstance(purpose, str) else purpose
        if len(set(self.purpose)) < len(self.purpose):
            raise ValueError("Duplicate purposes found in the list. Please provide unique purposes.")
        if not _data_downloaded() or not _check_data_integrity():
            download_oxford_flowers_data()
        self.labels = self._load_labels(_IMAGE_LABEL_FILE)
        self.image_paths = self._load_image_paths()
        self.train_ids, self.val_ids, self.test_ids = self._load_set_ids(_SETID_FILE)
        self.image_paths, self.labels = self._filter_by_purpose()

    def _load_labels(self, labels_file: str) -> list[int]:
        """
        Load image labels from the given .mat file.

        :param labels_file: Path to the .mat file with labels.
        :return: List of labels.
        """
        mat_data = scipy.io.loadmat(labels_file)
        return mat_data['labels'].squeeze().tolist()

    def _load_image_paths(self) -> list[str]:
        """
        Get sorted paths to all images in the directory.

        :return: List of sorted image file paths.
        """
        images = sorted(
            [f for f in os.listdir(_IMAGE_DIR) if f.endswith('.jpg')]
        )
        return [os.path.join(_IMAGE_DIR, img) for img in images]

    def _load_set_ids(self, set_id_file: str) -> tuple[list[int], list[int], list[int]]:
        """
        Load train, validation, and test IDs from the setid.mat file.

        :param set_id_file: Path to the .mat file with set IDs.
        :return: Tuple of train, validation, and test IDs.
        """
        mat_data = scipy.io.loadmat(set_id_file)
        train_ids = mat_data['tstid'].squeeze().tolist() # Swaps train and test, since test contains significantly more images
        val_ids = mat_data['valid'].squeeze().tolist()
        test_ids = mat_data['trnid'].squeeze().tolist()
        return train_ids, val_ids, test_ids

    def _filter_by_purpose(self) -> tuple[list[str], list[int]]:
        """
        Filter images and labels based on the dataset purpose.

        :return: Filtered image paths and labels.
        """
        chosen_ids = []
        for p in self.purpose:
            match p:
                case 'train':
                    chosen_ids += self.train_ids
                case 'validation':
                    chosen_ids += self.val_ids
                case 'test':
                    chosen_ids += self.test_ids
                case _:
                    raise ValueError(f"Unknown purpose: {p}. Must be 'train', 'validation', or 'test'.")

        chosen_ids = list(set(chosen_ids))

        filtered_paths = [self.image_paths[i - 1] for i in chosen_ids]
        filtered_labels = [self.labels[i - 1] for i in chosen_ids]
        return filtered_paths, filtered_labels

    def __len__(self) -> int:
        """
        Get the total number of images in the dataset.

        :return: Length of the dataset.
        """
        return len(self.image_paths)

    def __getitem__(self, idx: int) -> tuple[np.ndarray, int, str]:
        """
        Get an image and its corresponding label.

        :param idx: Index of the image.
        :return: Tuple of transformed image, label, and image path.
        """
        img_path = self.image_paths[idx]
        label = self.labels[idx] if self.labels else -1

        image = cv2.cvtColor(cv2.imread(img_path), cv2.COLOR_BGR2RGB)

        if self.transform:
            image = self.transform(image)

        return image, label, img_path




import abc
from collections.abc import Iterator, Iterable, MutableSequence
from functools import lru_cache, wraps
import warnings
from typing import Callable, Optional, Any

import cv2
import numpy as np
from sklearn.decomposition import PCA
from sklearn.cluster import KMeans

from .._config import setup_logging
from ..features._features import FeatureExtractorBase
from .._base_classes import SimilarityMetric

setup_logging()


def check_desired_output(
    similarity_func: Callable[[np.ndarray, np.ndarray], Any],
    vecs1: np.ndarray,
    vecs2: np.ndarray,
) -> Callable[[np.ndarray, np.ndarray], np.ndarray]:
    """
    Checks the output of the given similarity_func(vecs1, vecs2).
    Requirements:
    1) Output must be a NumPy array
    2) Output shape must be (len(vecs1), len(vecs2)) if batch
       or (1,1) if single
    3) If it fails, we degrade to a fallback method that
       loops over each row in vecs1 vs each row in vecs2.

    :param similarity_func: function that tries to compute similarities
                           between two arrays of shape (N, D) and (M, D).
    :param vecs1: (N, D) or (D,) array
    :param vecs2: (M, D) or (D,) array
    :return: A potentially wrapped function that always returns
             shape (N, M) as a NumPy array of floats
    """
    try:
        out = similarity_func(vecs1, vecs2)
    except Exception as e:
        warnings.warn(f"Similarity function threw an error: {e}. Falling back to row-wise loop.")
        return _make_fallback_func(similarity_func)

    if not isinstance(out, np.ndarray):
        warnings.warn(f"Expected a NumPy array, got {type(out)}. Using fallback method.")
        return _make_fallback_func(similarity_func)

    # Check shape
    # If vecs1 is shape (N, D) and vecs2 is shape (M, D), we expect out.shape = (N, M).
    # If single vector, it might produce shape (1,1) or just a float
    shape_ok = True
    if out.ndim == 2:
        if out.shape[0] != vecs1.shape[0] or out.shape[1] != vecs2.shape[0]:
            shape_ok = False
    elif out.ndim == 1 and out.size != 1:
        shape_ok = False

    if not shape_ok:
        warnings.warn(f"Output shape {out.shape} is not the expected (N, M). Expected output shape to be "
                                f"({vecs1.shape[0]}, {vecs2.shape[0]}). Using fallback.")
        return _make_fallback_func(similarity_func)

    return similarity_func

def _make_fallback_func(sim_func: Callable[[np.ndarray, np.ndarray], Any]) -> Callable[[np.ndarray, np.ndarray], np.ndarray]:
    """
    Returns a new function that loops row-by-row if the original
    similarity function can't handle batch mode.
    """
    def fallback(vecs1: np.ndarray, vecs2: np.ndarray) -> np.ndarray:
        # shape: (N, D) and (M, D)
        N = vecs1.shape[0]
        M = vecs2.shape[0]
        out = np.zeros((N, M), dtype=np.float32)
        for i in range(N):
            for j in range(M):
                out[i, j] = sim_func(vecs1[i:i+1], vecs2[j:j+1])
        return out
    try:
        return fallback
    except Exception as e:
        raise ValueError(f"Row-wise operation was not possible with the given similarity function: {e}"
                         "Your function is invalid.")

def _tupleize_first_arg(func: Callable) -> Callable:
    """
    # TODO: currently, the param 'image_paths' param is hardcoded. This should be more general
    # TODO: to be able to handle any variable name
    Pass this to the "encode" and "generate_encoding_map" methods to
    convert the input to a tuple so that it can be hashed by the lru_cache.
    """
    @wraps(func)
    def wrapper(self, image_paths: Any, /, *args, **kwargs):
        if isinstance(image_paths, (Iterator, MutableSequence)):
            image_paths = tuple(image_paths)
        return func(self, image_paths, *args, **kwargs)
    return wrapper

class ImageEncoderBase(SimilarityMetric):
    """
    Base class for image encoders. An image encoder is a class that
    generates a vector representation of an image. Subclasses use a combination of:

    - A feature extractor: Extract local features from an image (e.g. SIFT, SURF, or Deep Features).
    - A clustering model (K-Means for VLAD or GMM for Fisher Vector): aggregates local features to their
    nearest centroids to produce fix-sized vectors.
    - A similarity function: computes a single float value from the vector representations that represents
    the similarity between two images.

    The encoding can be used for indexing, retrieval, clustering or classification tasks.
    :param feature_extractor: Feature extractor instance (should implement __call__).
    :param clustering_model: Clustering model used for generating descriptors.
    :param power_norm_weight: Exponent for power normalization
    :param norm_order: Norm order for normalization (default: 2).
    :param epsilon: Small constant to avoid division by zero.
    :param flatten: Whether to flatten the computed descriptor vector (default: True).
    :param similarity_func: A function that takes two batches of vectors and returns a similarity score
    matrix with size (batch_1_size, batch_2_size).
    :param pca: PCA model for dimensionality reduction (optional).
    :param raise_error_when_pca_incompatible: When set to True, if the new clustering model has a different input size
                                        than the PCA model's output size, an Error will be raised """
    def __init__(
            self,
            feature_extractor: FeatureExtractorBase=None,
            clustering_model=None,
            similarity_func: Callable[[np.ndarray, np.ndarray], float]=None,
            power_norm_weight: float=1,
            norm_order: int = 2,
            epsilon: float = 1e-9,
            flatten: bool = True,
            pca: Optional[PCA] = None,
            raise_error_when_pca_incompatible: bool = True
    ):
        # Set important attributes via setters to trigger error handling
        self._feature_extractor = None
        self._clustering_model = None
        self._pca = None
        self._similarity_func = None

        self.similarity_func = similarity_func
        if clustering_model is not None:
            self.clustering_model = clustering_model
        if pca is not None:
            self.pca = pca

        self.feature_extractor = feature_extractor

        self.power_norm_weight = power_norm_weight
        self.norm_order = norm_order
        self.epsilon = epsilon
        self.flatten = flatten
        self.raise_error_when_pca_incompatible = raise_error_when_pca_incompatible

    @property
    def feature_extractor(self) -> FeatureExtractorBase:
        return self._feature_extractor

    @feature_extractor.setter
    def feature_extractor(self, feature_extractor: FeatureExtractorBase):
        if not isinstance(feature_extractor, FeatureExtractorBase):
            raise TypeError(f"feature_extractor must be an instance of FeatureExtractorBase, not {type(feature_extractor)}")
        if self._pca:
            if feature_extractor.output_dim != self._pca.n_features_in_:
                raise RuntimeError(f"Feature Extractor outputs shape {feature_extractor.output_dim}, "
                                   f"But PCA accepts input dim {self._pca.n_features_in_}")
        else:
            if self._clustering_model is not None:
                if feature_extractor.output_dim != self._clustering_model.n_features_in_:
                    raise RuntimeError(f"Feature Extractor outputs shape {feature_extractor.output_dim}, "
                                       f"But clustering model accepts input dim {self._clustering_model.n_features_in_}")
        self._feature_extractor = feature_extractor

    @property
    def similarity_func(self):
        return self._similarity_func

    @similarity_func.setter
    def similarity_func(self, func: Callable[[np.ndarray, np.ndarray], float]):
        dummy1, dummy2 = np.random.rand(10, 10), np.random.rand(10, 10)
        self._similarity_func = check_desired_output(func, dummy1, dummy2)

    @property
    def clustering_model(self):
        return self._clustering_model

    @clustering_model.setter
    def clustering_model(self, clustering_model):
        if self._feature_extractor.output_dim != self._clustering_model.n_features_in_:
            raise RuntimeError(f"Feature Extractor outputs shape {self.feature_extractor.output_dim}, "
                               f"But clustering model accepts input dim {self._clustering_model.n_features_in_}")
        self._clustering_model = clustering_model

        if self._pca:
            if self._pca.n_components != self._clustering_model.n_features_in_:
                if self.raise_error_when_pca_incompatible:
                    raise RuntimeError(f"PCA is incompatible with the new clustering model. "
                                    f"PCA input size: {self._pca.n_components}, "
                                    f"New clustering model input size: {self._clustering_model.n_features_in_}. ")
                warnings.warn(f"PCA is incompatible with the new clustering model. "
                              f"PCA input size: {self._pca.n_components}, "
                              f"New clustering model input size: {self._clustering_model.n_features_in_}. "
                              "PCA will be reset to None to avoid errors."
                              "If you want to raise an Error instead when this happens, set raise_error_when_pca_incompatible=False.")
                self._pca = None

    @property
    def pca(self):
        return self._pca

    @pca.setter
    def pca(self, value: PCA):
        if not self._clustering_model:
            raise ValueError("PCA cannot be set without an existing clustering model.")

        if value.n_components != self._clustering_model.n_features_in_:
            raise ValueError("PCA input size has to match the clustering model input size."
                             f"PCA model has input size {value.n_components}, "
                             f"while clustering model has input size {self._clustering_model.n_features_in_}")

        self._pca = value

    def learn(self, images: Iterable[np.ndarray], /,*, n_clusters: int, reduced_dim: bool = False, reduce_factor: int=2, **kwargs) -> None:
        """
        Learns the visual vocabulary from a collection of images.

        :param images: An iterable of images. First argument of the iterable is expected to be the image itself.
        :param n_clusters: Number of clusters to use for the clustering model
        :param reduced_dim: Whether to apply PCA for dimensionality reduction
        :param reduce_factor: Factor to reduce the dimension by
        :param kwargs: Additional arguments for the clustering model
        """
        features = np.vstack([self.feature_extractor(image) for image, *_ in images])
        if reduced_dim:
            self._pca = PCA(n_components=features.shape[1] // reduce_factor)
            self._pca.fit(features)
            features = self._pca.transform(features)
        clustering_model = KMeans(n_clusters=n_clusters, **kwargs)
        clustering_model.fit(features)
        self.clustering_model = clustering_model

    @_tupleize_first_arg
    @lru_cache(maxsize=4)
    def generate_encoding_map(self, image_paths: Iterable[str], /) -> dict[str, np.ndarray]:
        """
        Converts a collection of image file paths into a dictionary of
        ``{image_path: encoded_vector}``.

        This method automatically reads each image, applies the internal
        encoding pipeline, and stores the resulting descriptor vector.

        :param image_paths: List of image full paths
        :return: a dictionary where keys are image paths and values are descriptor vectors of the
                corresponding images
        """
        images = (cv2.cvtColor(cv2.imread(path), cv2.COLOR_BGR2RGB) for path in image_paths)
        return dict(zip(image_paths, self.encode(images)))

    @abc.abstractmethod
    def encode(self, images: Iterable[np.ndarray] | np.ndarray) -> np.ndarray:
        """
        Encodes one or more images into a batch of vector representations.

        :param images: iterable. Consider using an iterator if you have a lot of images.
        :return: vector representations of the given images
        """
        raise NotImplementedError

    def similarity_score(self, images1: Iterable[np.ndarray] | np.ndarray, images2: Iterable[np.ndarray] | np.ndarray) -> float:
        """
        Computes vector encodings for two images and calculates the similarity score between them.

        :param images1: First (batch of) image(s)
        :param images2: Second (batch of) image(s)
        :return: Similarity score. If image iterables are provided, a similarity matrix between two image batches is returned.
        """
        vector1 = self.encode(images1)
        vector2 = self.encode(images2)
        result = self.similarity_func(vector1, vector2)
        return np.float_(result)

    def __repr__(self) -> str:
        n_clusters = None
        if self._clustering_model:
            if hasattr(self._clustering_model, 'n_clusters'):
                n_clusters = self._clustering_model.n_clusters
            elif hasattr(self._clustering_model, 'n_components'):
                n_clusters = self._clustering_model.n_components
        return self.__class__.__name__ + f"(feature_extractor={self.feature_extractor.__class__.__name__}, \n" \
               f"similarity_func={self.similarity_func.__name__}, \n" \
               f"Number of clusters/components={n_clusters}, \n" \
               f"power_norm_weight={self.power_norm_weight}, \n" \
               f"norm_order={self.norm_order})"


"""
Includes exceptions for the package.
"""

class InvalidImageError(Exception):
    """
    Raised when an image is not provided.
    """
    def __init__(self, message: str = "Input is not a valid image."):
        super().__init__(message)


from typing import Callable, Iterable

import numpy as np
import torch
from sklearn.decomposition import PCA
from sklearn.mixture import GaussianMixture

from ..features import RootSIFT
from ..features._features import FeatureExtractorBase
from ..encoders._base_encoder import ImageEncoderBase
from .._utils import cosine_similarity


class FisherVectorEncoder(ImageEncoderBase):
    """
    This class serves as an encoder that transforms input images into Fisher Vector descriptors.

    The Fisher Vector representation is based on the gradients of the GMM parameters
    (weights, means, and covariances) with respect to the feature descriptors extracted
    from the images. The representation is optionally power-normalized and L2-normalized.

    The output when calling `compute_vector` has shape (2 * num_clusters * feature_dim + num_clusters,).

    :param feature_extractor: Feature extractor instance. Default is RootSIFT
    :param gmm_model: Gaussian Mixture Model instance from scikit-learn.
    :param power_norm_weight: Exponent for power normalization
    :param norm_order: Norm order for normalization (default: 2).
    :param epsilon: Small constant to avoid division by zero.
    :param flatten: Whether to flatten the computed encoding vector (default: True).
    :param similarity_func: A function that takes two batches of vectors and returns a similarity score
    matrix with size (batch_1_size, batch_2_size).
    :param pca: PCA model for dimensionality reduction (optional).
    :param raise_error_when_pca_incompatible: When set to True, if the new clustering model has a different input size
                                        than the PCA model's output size, the PCA model will be reset to None.

    References:
    ==========
    [1] Hervé Jégou, Florent Perronnin, Matthijs Douze, Jorge Sánchez, Patrick Pérez, and Cordelia Schmid, "Aggregating Local Image Descriptors into Compact Codes," IEEE.
    """
    def __init__(self,
                 feature_extractor: FeatureExtractorBase=RootSIFT(),
                 gmm_model: GaussianMixture=None,
                 power_norm_weight: float = 0.5,
                 norm_order: int = 2,
                 epsilon: float = 1e-9,
                 flatten: bool = True,
                 similarity_func: Callable[[np.ndarray, np.ndarray], float] = cosine_similarity,
                 pca: PCA = None,
                 raise_error_when_pca_incompatible: bool = False):
        if gmm_model is not None:
            if not isinstance(gmm_model, GaussianMixture):
                raise ValueError(f"The clustering model must be an instance of GaussianMixture, not {type(gmm_model)}")
        super().__init__(feature_extractor,
                         gmm_model,
                         similarity_func,
                         power_norm_weight,
                         norm_order,
                         epsilon,
                         flatten,
                         pca,
                         raise_error_when_pca_incompatible)

    def encode(self, images: Iterable[np.ndarray] | np.ndarray) -> np.ndarray:
        all_encodings = []
        if isinstance(images, torch.Tensor):
            raise RuntimeError("Torch images are not supported yet.")
        if isinstance(images, np.ndarray) and images.ndim == 3:
            images = [images] # Handle single image case
        for image in images:
            descriptors = self.feature_extractor(image)
            if self.pca:
                descriptors = self.pca.transform(descriptors.astype(np.float32))
            num_descriptors = len(descriptors)

            mixture_weights = self.clustering_model.weights_
            means = self.clustering_model.means_
            covariances = self.clustering_model.covariances_

            posterior_probabilities = self.clustering_model.predict_proba(descriptors)

            # Statistics necessary to compute GMM gradients wrt its parameters
            pp_sum = posterior_probabilities.mean(axis=0, keepdims=True).T
            pp_x = posterior_probabilities.T.dot(descriptors) / num_descriptors
            pp_x_2 = posterior_probabilities.T.dot(np.power(descriptors, 2)) / num_descriptors

            # Compute GMM gradients wrt its parameters
            d_pi = pp_sum.squeeze() - mixture_weights

            d_mu = pp_x - pp_sum * means

            d_sigma_t1 = pp_sum * np.power(means, 2)
            d_sigma_t2 = pp_sum * covariances
            d_sigma_t3 = 2 * pp_x * means
            d_sigma = -pp_x_2 - d_sigma_t1 + d_sigma_t2 + d_sigma_t3

            # Apply analytical diagonal normalization
            sqrt_mixture_weights = np.sqrt(mixture_weights)
            d_pi /= sqrt_mixture_weights
            d_mu /= sqrt_mixture_weights[:, np.newaxis] * np.sqrt(covariances)
            d_sigma /= np.sqrt(2) * sqrt_mixture_weights[:, np.newaxis] * covariances

            # Concatenate GMM gradients to form Fisher vector representation
            descriptor_vector = np.hstack((d_pi, d_mu.ravel(), d_sigma.ravel()))
            descriptor_vector = descriptor_vector.reshape(1, -1)

            # Power normalization and L2 normalization
            descriptor_vector = np.sign(descriptor_vector) * np.power(np.abs(descriptor_vector), self.power_norm_weight)
            norm = np.linalg.norm(descriptor_vector, axis=1, ord=self.norm_order, keepdims=True) + self.epsilon
            descriptor_vector = descriptor_vector / norm

            if self.flatten:
                descriptor_vector = descriptor_vector.flatten()
            all_encodings.append(descriptor_vector)

        return np.vstack(all_encodings)



import logging
import logging.config
import pathlib

import yaml

# -Config for the dataset- #
ROOT = pathlib.Path(__file__).parent.parent
LOG_FILE_PATH = ROOT / "res/logs/log_msgs.log"

# - Logging - #
def setup_logging(default_path=rf"{ROOT}/res/logging_config.yaml", default_level=logging.INFO):
    """Setup logging configuration"""
    try:
        with open(default_path, 'rt') as f:
            config = yaml.safe_load(f.read())
        try:
            config["handlers"]["file_handler"]["filename"] = str(LOG_FILE_PATH)
        except Exception as e:
            print(f"Error in Logging Configuration: {e}. Cannot set output path for log file.")
        logging.config.dictConfig(config)
    except Exception as e:
        print(f"Error in Logging Configuration: {e}")
        logging.basicConfig(level=default_level, format='%(asctime)s - %(name)s - %(levelname)s - %(message)s')





"""
This module contains functions to evaluate the performance of a retrieval system.
"""
from itertools import islice
from typing import Iterable

import numpy as np

from ._utils import *

__all__ = ['retrieve_top_k_similar', 'top_k_map', 'top_k_accuracy']

def retrieve_top_k_similar(
    uploaded_image: np.ndarray,
    dataset: dict[str, np.ndarray],
    encoder,
    k: int = 5
) -> list[tuple[str, float]]:
    """
    Returns the top-k most similar images from 'dataset' to the 'uploaded_image'.

    :param uploaded_image: Query image as a NumPy array (H x W x C).
    :param dataset: A dict mapping file paths to their feature vectors (np.ndarray).
    :param encoder: An object that implements `compute_vector(img) -> np.ndarray`.
    :param k: Number of top similar images to return.
    :return: A list of (image_path, similarity_score) for the top-k matches, sorted descending by similarity.
    """
    all_vectors, all_paths = np.array(list(dataset.values())), list(dataset.keys())

    # Query vector
    query_vector = encoder.encode(uploaded_image)

    # If `query_vector` is 1D, reshape to (1, D) to work with `cosine_similarity`
    if query_vector.ndim == 1:
        query_vector = query_vector.reshape(1, -1)

    scores = cosine_similarity(query_vector, all_vectors)   # (1, N)
    scores = scores[0]

    sorted_indices = np.argsort(-scores)  # highest scores first

    # Slice top-k
    top_k_indices = sorted_indices[:k]

    results = [(all_paths[i], scores[i]) for i in top_k_indices]
    return results


def top_k_map(images: Iterable[np.ndarray],
              image_labels: Iterable[int],
              encoding_map: dict[str, np.ndarray],
              path_labels_dict: dict[str, int],
              encoder,
              k: int=None) -> float:
    """
    Computes mean Average Precision over the queries,
    based on whether retrieved images have matching labels.

    :param images: List of query images (NumPy arrays).
    :param image_labels: Corresponding labels for the query images.
    :param encoding_map: dict {img_path: feature_vector}
    :param path_labels_dict: dict {img_path: label}
    :param encoder: Object with `compute_vector(img)
    :param k: Number of top results to consider.
    :return: mAP
    """
    all_vectors, all_paths = np.array(list(encoding_map.values())), list(encoding_map.keys())

    APs = []
    for query_img, true_label in zip(images, image_labels):
        query_vec = encoder.encode(query_img)
        if query_vec.ndim == 1:
            query_vec = query_vec.reshape(1, -1)

        sims = cosine_similarity(query_vec, all_vectors)[0]

        # Sort by descending similarity
        sorted_idx = np.argsort(-sims)
        if k is not None:
            sorted_idx = sorted_idx[:k]

        sorted_paths = [all_paths[i] for i in sorted_idx]
        sorted_labels = [path_labels_dict[path] for path in sorted_paths]

        # compute average precision by counting relevant images at each rank
        relevant_count = 0
        precision_sum = 0.0
        for rank, path in enumerate(sorted_paths, start=1):
            if path_labels_dict[path] == true_label:
                relevant_count += 1
                precision_sum += (relevant_count / rank)

        # If there are R relevant images in the entire dataset
        # average precision = sum(precision_at_i for each relevant i) / R
        R = sum(lbl == true_label for lbl in sorted_labels)
        AP = precision_sum / R if R > 0 else 0.0

        APs.append(AP)

    return float(np.mean(APs))

def top_k_accuracy(
    images: Iterable[np.ndarray],
    image_labels: Iterable[int],
    encoding_map: dict[str, np.ndarray],
    path_labels_dict: dict[str, int],
    encoder,
    k: int
) -> float:
    """
    Computes top-k accuracy. For each query, we look at the top-k
    most similar results in the dataset. If any of them match the
    query's label, that query is considered correct.

    :param images: List of query images.
    :param image_labels: List of true labels for each query image.
    :param encoding_map: dict {path: feature_vector}.
    :param path_labels_dict: dict {path: label}.
    :param encoder: An object with `compute_vector(img) -> np.ndarray`.
    :param k: Number of top results to check for a correct match.
    :return: Top-k accuracy (float) in the range [0, 1].
    """
    all_paths, all_vectors = list(encoding_map.keys()), np.array(list(encoding_map.values()))
    correct_count = 0

    for query_img, true_label in zip(images, image_labels):
        q_vec = encoder.encode(query_img)
        if q_vec.ndim == 1:
            q_vec = q_vec.reshape(1, -1)

        sims = cosine_similarity(q_vec, all_vectors)[0]
        sorted_idx = np.argsort(-sims)[:k]               # top-k

        # Check if any of the top-k share the query's label
        found_match = False
        for idx in sorted_idx:
            if path_labels_dict[all_paths[idx]] == true_label:
                found_match = True
                break

        if found_match:
            correct_count += 1

    topk_acc = correct_count / len(images)
    return float(topk_acc)



"""
TODO: This module was made to compute loss of segmentation model_files.
TODO: For the future, implement loss for retrieval (e.g. triplet loss, contrastive loss, "similarity loss" etc.)
"""

import logging
from typing import Optional

import torch
from torch.nn.modules.loss import _Loss
import torch.nn.functional as F

from pyvisim._utils import soft_dice_score

__all__ = ["MultiClassDiceLoss"]

class MultiClassDiceLoss(_Loss):
    __name__ = "MultiClassDiceLoss"
    def __init__(
                self,
                mode: str,
                classes: Optional[torch.Tensor] = None,
                log_loss: bool = False,
                from_logits: bool = True,
                smooth: float = 0.0,
                eps: float = 1e-7,
                ignore_index: Optional[int] = None) -> None:
        """
        Dice loss for image segmentation task.
        """
        super().__init__()
        assert mode in {'binary', 'multiclass'}, f"Unknown mode: {mode}. Supported modes are 'multiclass' and 'binary'."
        self.mode = mode
        self.classes = classes
        self.from_logits = from_logits
        self.smooth = smooth
        self.eps = eps
        self.log_loss = log_loss
        self.ignore_index = ignore_index

    def forward(self, y_pred: torch.Tensor, y_true: torch.Tensor) -> torch.Tensor:
        """
        Forward pass.

        :param y_pred: (B, C, H, W)
        :param y_true: (B, C, H, W)

        :return: dice loss
        """
        assert y_true.dim() == y_pred.dim() == 4, f"Expected 4D input tensors, got {y_pred.dim()} for y_pred and {y_true.dim()} for y_true"
        logging.debug(f"""
                    Batch size: {y_true.size(0)}
                    Number of classes: {y_true.size(1)}
                    Image shape: ({y_true.size(2)} x {y_true.size(3)})
                    """)
        if self.from_logits:
            if self.mode == 'multiclass':
                y_pred = F.softmax(y_pred, dim=1)
            elif self.mode == 'binary':
                y_pred = F.sigmoid(y_pred)

        batch_size = y_true.size(0)
        num_classes = y_pred.size(1)
        dims = (0, 2)

        y_true = y_true.view(batch_size, num_classes, -1)
        y_pred = y_pred.view(batch_size, num_classes, -1)

        if self.ignore_index is not None:
            mask = y_true != self.ignore_index
            y_pred = y_pred * mask

        scores = soft_dice_score(y_pred, y_true.type_as(y_pred), smooth=self.smooth, eps=self.eps, dims=dims)

        if self.log_loss:
            loss = -torch.log(scores.clamp_min(self.eps))
        else:
            loss = 1.0 - scores

        mask = y_true.sum(dims) > 0 # Zeros out loss of classes that are not present in the mask (otherwise they would have a loss of 1, which is nonsense!
        loss *= mask.to(loss.dtype)

        if self.classes is not None:
            loss = loss[self.classes]

        return loss.mean()


class FocalLoss(_Loss):
    __name__ = "FocalLoss"
    def __init__(self,
                 mode: str,
                 alpha: Optional[torch.Tensor] = None,
                 normalize_weights: bool = True,
                 gamma: float = 2.0,
                 from_logits: bool = True,
                 ignore_index: Optional[int] = None) -> None:
        """
        Focal loss for image segmentation task.
        """
        super().__init__()
        assert mode in {'binary', 'multiclass'}, f"Unknown mode: {mode}. Supported modes are 'multiclass' and 'binary'."
        self.mode = mode
        self.alpha = alpha
        if self.alpha is not None and normalize_weights:
            self.alpha = self.alpha / self.alpha.sum()

        self.gamma = gamma
        self.from_logits = from_logits
        self.ignore_index = ignore_index

    def forward(self, y_pred: torch.Tensor, y_true: torch.Tensor) -> torch.Tensor:
        """
        Forward pass.

        :param y_pred: (B, C, H, W)
        :param y_true: (B, C, H, W)

        :return: Weighted focal loss
        """
        assert y_pred.dim() == y_pred.dim() == 4, f"Expected 4D input tensors, got {y_pred.dim()} for y_pred and {y_true.dim()} for y_true"
        y_true = torch.argmax(y_true, dim=1) # Revert one-hot encoding back to (B, H, W) tensor

        if self.from_logits:
            if self.mode == 'multiclass':
                y_pred = F.softmax(y_pred, dim=1)
            elif self.mode == 'binary':
                y_pred = torch.sigmoid(y_pred)

        if self.mode == 'multiclass':
            num_classes = y_pred.size(1)
            # Flatten predictions and targets
            y_pred = y_pred.permute(0, 2, 3, 1).reshape(-1, num_classes)  # Shape: (N, C)
            y_true = y_true.view(-1)  # Shape: (N,)

            if self.ignore_index is not None:
                valid_mask = y_true != self.ignore_index
                y_pred = y_pred[valid_mask]
                y_true = y_true[valid_mask]

            # Gather probabilities of the true class
            p_t = y_pred[torch.arange(y_pred.size(0)), y_true]

            # Handle alpha
            if self.alpha is None:
                self.alpha = torch.ones(num_classes, device=y_pred.device)/num_classes
            if not self.alpha.device == y_pred.device:
                self.alpha = self.alpha.to(y_pred.device)
            alpha_t = self.alpha[y_true]

        elif self.mode == 'binary':
            y_pred = y_pred.view(-1)
            y_true = y_true.view(-1).float()

            if self.ignore_index is not None:
                valid_mask = y_true != self.ignore_index
                y_pred = y_pred[valid_mask]
                y_true = y_true[valid_mask]

            p_t = y_pred * y_true + (1 - y_pred) * (1 - y_true)
            if self.alpha is not None:
                alpha_t = self.alpha * y_true + (1 - self.alpha) * (1 - y_true)
            else:
                alpha_t = 1.0

        # Compute focal loss
        focal_weight = alpha_t * (1 - p_t) ** self.gamma
        loss = focal_weight * (-torch.log(p_t.clamp(min=1e-7)))

        return loss.mean()

class HybridFocalDiceLoss(_Loss):
    def __init__(self,
                 mode: str,
                 alpha: Optional[torch.Tensor] = None,
                 gamma: float = 2.0,
                 from_logits: bool = True,
                 ignore_index: Optional[int] = None,
                 dice_weight: float = 1.0,
                 focal_weight: float = 1.0,
                 smooth: float = 1e-5,
                 eps: float = 1e-7) -> None:
        super().__init__()
        self.focal_loss = FocalLoss(mode=mode,
                                    alpha=alpha,
                                    gamma=gamma,
                                    from_logits=from_logits,
                                    ignore_index=ignore_index)
        self.dice_loss = MultiClassDiceLoss(mode=mode,
                                            from_logits=from_logits,
                                            smooth=smooth,
                                            eps=eps)
        if not dice_weight + focal_weight == 1.0:
            raise ValueError(f"Sum of dice_weight and focal_weight must be equal to 1.0, got {dice_weight} + {focal_weight} = {dice_weight + focal_weight}")
        self.dice_weight = dice_weight
        self.focal_weight = focal_weight

    def forward(self, y_pred: torch.Tensor, y_true: torch.Tensor) -> torch.Tensor:
        focal = self.focal_loss(y_pred, y_true)
        dice = self.dice_loss(y_pred, y_true)
        loss = self.focal_weight * focal + self.dice_weight * dice
        return loss





from typing import Callable, Iterable

import numpy as np
import torch
from sklearn.cluster import KMeans
from sklearn.decomposition import PCA

from ..features._features import FeatureExtractorBase
from ..encoders._base_encoder import ImageEncoderBase
from .._utils import cosine_similarity


class VLADEncoder(ImageEncoderBase):
    """
    This class encodes images into VLAD descriptor vectors
    using a chosen feature extractor and a pretrained K-Means model,
    then compares two VLAD descriptor vectors with a user-specified
    or default (cosine) similarity function.

    The output when calling `compute_vector` has shape (num_clusters * feature_dim,).

    You can use euclidean distance, manhattan distance, etc. as the similarity function.

    The encoding can be used for indexing, retrieval, clustering or classification tasks.
    :param feature_extractor: Feature extractor instance (should implement __call__).
    :param kmeans_model: KMeans model instance from scikit-learn.
    :param power_norm_weight: Exponent for power normalization
    :param norm_order: Norm order for normalization (default: 2).
    :param epsilon: Small constant to avoid division by zero.
    :param flatten: Whether to flatten the computed descriptor vector (default: True).
    :param similarity_func: A function that takes two batches of vectors and returns a similarity score
    matrix with size (batch_1_size, batch_2_size).
    :param pca: PCA model for dimensionality reduction (optional).
    :param raise_error_when_pca_incompatible: When set to True, if the new clustering model has a different input size
                                        than the PCA model's output size, the PCA model will be reset to None.

    References:
    ==========
    [1] Relja Arandjelović and Andrew Zisserman, 'All About VLAD', Department of Engineering Science, University of Oxford.
    [2] Relja Arandjelović and Andrew Zisserman, "Three things everyone should know to improve object retrieval," Department of Engineering Science, University of Oxford.
    [3] Hervé Jégou, Florent Perronnin, Matthijs Douze, Jorge Sánchez, Patrick Pérez, and Cordelia Schmid, "Aggregating Local Image Descriptors into Compact Codes," IEEE.
    """
    def __init__(
            self,
            feature_extractor: FeatureExtractorBase,
            kmeans_model: KMeans=None,
            power_norm_weight: float = 1, # no paper found where power norm weight is used for VLAD
            norm_order: int = 2,
            epsilon: float = 1e-9,
            flatten: bool = True,
            similarity_func: Callable[[np.ndarray, np.ndarray], float] = cosine_similarity,
            pca: PCA = None,
            raise_error_when_pca_incompatible: bool = False) -> None:
        if kmeans_model is not None:
            if not isinstance(kmeans_model, KMeans):
                raise ValueError(f"The clustering model must be an instance of KMeans, not {type(kmeans_model)}")
        super().__init__(feature_extractor,
                         kmeans_model,
                         similarity_func,
                         power_norm_weight,
                         norm_order,
                         epsilon,
                         flatten,
                         pca,
                         raise_error_when_pca_incompatible)

    def encode(self, images: Iterable[np.ndarray] | np.ndarray) -> np.ndarray:
        all_encodings = []
        if isinstance(images, torch.Tensor):
            raise RuntimeError("Torch images are not supported yet.")
        if isinstance(images, np.ndarray) and images.ndim == 3:
            images = [images] # Handle single image case
        for image in images:
            descriptors = self.feature_extractor(image)
            if self.pca:
                descriptors = self.pca.transform(descriptors.astype(np.float32))

            if descriptors is None or descriptors.shape[0] == 0:
                return np.zeros(len(self.clustering_model.cluster_centers_) * descriptors.shape[1], dtype=np.float32)

            labels = self.clustering_model.predict(descriptors.astype(np.float32))
            centroids = self.clustering_model.cluster_centers_

            k = len(centroids)
            dim = descriptors.shape[1]
            descriptor_vector = np.zeros((k, dim), dtype=np.float32)

            for i, desc in enumerate(descriptors):
                cluster_id = labels[i]
                descriptor_vector[cluster_id] += (desc - centroids[cluster_id])

            descriptor_vector = np.sign(descriptor_vector) * np.abs(descriptor_vector) ** self.power_norm_weight
            norms = np.linalg.norm(descriptor_vector, axis=1, ord=self.norm_order, keepdims=True) + self.epsilon
            descriptor_vector = descriptor_vector / norms

            if self.flatten:
                descriptor_vector = descriptor_vector.flatten()

            all_encodings.append(descriptor_vector)

        return np.vstack(all_encodings)



import io
import json
import logging
import math
import os
import shutil
from dataclasses import dataclass
from enum import Enum
from typing import Type, Optional, Any, Union, Sequence, Literal

import cv2
import h5py
import joblib
import matplotlib.pyplot as plt
import numpy as np
import seaborn as sns
import torch
import torchvision.transforms.functional as TF
from scipy.stats import pearsonr, spearmanr
from segmentation_models_pytorch.utils.metrics import IoU
from sklearn.cluster import KMeans, DBSCAN, SpectralClustering
from sklearn.metrics import adjusted_rand_score, adjusted_mutual_info_score, rand_score
from sklearn.linear_model import LinearRegression
from sklearn.metrics import silhouette_score, mean_squared_error
from sklearn.metrics.pairwise import euclidean_distances, cosine_similarity as cs
from tqdm import tqdm

from ._config import setup_logging
from ._errors import InvalidImageError

setup_logging()


# Decorators
def check_is_image(arg_positions: tuple=None, kwarg_positions: tuple=None, tol=1e-5):
    """
    Decorator to check if one or more arguments are valid images. Default is the first positional
    argument.

    **Note**: both 'arg_positions' abd 'kwarg_positions' are zero-based!

    :param arg_positions: A tuple of positions (integers) in *args
                          that correspond to the position of the images / masks.
    :param kwarg_positions: A tuple of positions (integers) in **kwargs
                            that correspond to the position of the images / masks.
    :param tol: tolerance for float comparison for Torch images.
    """
    def decorator(func):
        def wrapper(*args, **kwargs):
            def generator(*gen_args, **gen_kwargs):
                if arg_positions:
                    for pos in arg_positions:
                        yield gen_args[pos]
                if kwarg_positions:
                    kw_vals = (val for i, val in enumerate(gen_kwargs.values()) if i in kwarg_positions)
                    for kw_val in kw_vals:
                        yield kw_val
                if not arg_positions and not kwarg_positions:
                    yield gen_args[0]
            for pos, image in enumerate(generator(*args, **kwargs)):
                if not hasattr(image, 'shape'):
                    raise InvalidImageError(f"Argument at position {pos} of type {type(image)} does not have attribute 'shape'. So it is neither a numpy array nor a torch tensor.")
                if not (2 <= len(image.shape) <= 3):
                    raise InvalidImageError(f"Image must be 2D or 3D. Got shape {image.shape} for position {pos}")

                if isinstance(image, np.ndarray):
                    if len(image.shape) == 2:
                        if not np.all(image == image.astype(np.int64)):
                            raise InvalidImageError(f"Mask values must be integers. Got min={image.min()} and max={image.max()}.")
                    else:
                        if image.shape[2] != 3:
                            raise InvalidImageError(f"NumPy 3D images must have shape (H, W, 3). Got {image.shape}.")
                        if image.min() < 0 or image.max() > 255:
                            raise InvalidImageError(f"Image values must be in the range [0, 255]. Got min={image.min()} and max={image.max()} for position {pos}")

                elif torch.is_tensor(image):
                    if len(image.shape) == 2:
                        if not torch.all(image == image.to(torch.int)):
                            raise InvalidImageError(f"Mask values must be integers. Got min={image.min().item()} and max={image.max().item()} for position {pos}")
                    else:
                        if image.shape[0] != 3:
                            raise InvalidImageError(f"Torch 3D images must have shape (3, H, W). Got {image.shape}.")
                        if image.min().item() < 0.0 - tol or image.max().item() > 1.0 + tol:
                            raise InvalidImageError(f"Image values must be in the range [0, 1]. Got min={image.min().item()} and max={image.max().item()} for position {pos}")
                else:
                    raise InvalidImageError(f"Input must be a numpy array or a torch tensor, not {type(image)}.")

            return func(*args, **kwargs)
        return wrapper
    return decorator


def cluster_and_return_labels(data: np.ndarray,
                              method: Literal['kmeans', 'dbscan', 'spectral'] = 'kmeans',
                              n_clusters: Optional[int] = None,
                              **kwargs) -> np.ndarray:
    """
    Clusters 'data' using the specified method.

    :param data: A 2D NumPy array of shape (N, D)
    :param method: 'kmeans', 'dbscan', or 'spectral'
    :param n_clusters: Number of clusters (if applicable)
    :param kwargs: Additional arguments to pass to the clustering constructor
    :return: 1D NumPy array of cluster labels (shape: (N,))
    """
    if method == 'kmeans':
        if n_clusters is None:
            raise ValueError("n_clusters must be specified for KMeans.")
        model = KMeans(n_clusters=n_clusters, random_state=42, **kwargs)
        labels = model.fit_predict(data)
        return labels

    elif method == 'dbscan':
        # DBSCAN doesn't need n_clusters (but can accept eps, min_samples)
        model = DBSCAN(**kwargs)
        labels = model.fit_predict(data)
        return labels

    elif method == 'spectral':
        if n_clusters is None:
            raise ValueError("n_clusters must be specified for Spectral Clustering.")
        model = SpectralClustering(n_clusters=n_clusters, affinity='nearest_neighbors', random_state=42, **kwargs)
        labels = model.fit_predict(data)
        return labels

    else:
        raise ValueError(f"Unknown method: {method}")


@check_is_image()
def plot_image(image: np.ndarray | torch.Tensor, title: str = "Image") -> None:
    """Simply olots the image."""
    plt.figure(figsize=(10, 10))
    if isinstance(image, torch.Tensor):
        image = image.permute(1, 2, 0)
    plt.imshow(image)
    plt.axis('off')
    plt.title(title)
    plt.show()


def save_json(file_path: str, data: dict) -> None:
    """
    Save the given data to a JSON file.

    :param file_path: Path to the JSON file
    :param data: Dictionary containing data to save
    """
    with open(file_path, 'w') as file:
        json.dump(data, file, indent=4)


def save_to_hdf5(file_path: str,
                 dataset_dict: dict[str, Any]) -> None:
    """
    Save data to an HDF5 file using concise match-case type handling.

    :param file_path: Path to the HDF5 file
    :param dataset_dict: Dictionary containing data to save

    :raises TypeError: If the data type is not supported
    """
    with h5py.File(file_path, 'w') as f:
        def save_to_hdf5_helper(dataset_dict: dict[str, Any], f: h5py.File):
            for dataset_name, data in dataset_dict.items():
                match data:
                    case int() | float():
                        f.create_dataset(dataset_name, data=data)

                    case torch.Tensor():
                        # Convert Torch tensor to NumPy array
                        data = data.numpy()
                        f.create_dataset(dataset_name, data=data)

                    case np.ndarray():
                        # Handle strings in NumPy arrays differently
                        if data.dtype.kind in {'U', 'S'}:  # Unicode or bytes
                            dt = h5py.string_dtype(encoding='utf-8')
                            f.create_dataset(dataset_name, data=data.astype(dt))
                        else:
                            f.create_dataset(dataset_name, data=data)

                    case list():
                        # Convert lists to NumPy arrays if possible
                        try:
                            np_data = np.array(data)
                            if np_data.dtype.kind in {'U', 'S'}:
                                dt = h5py.string_dtype(encoding='utf-8')
                                np_data = np_data.astype(dt)
                            f.create_dataset(dataset_name, data=np_data)
                        except ValueError as e:
                            raise ValueError(f"Cannot convert list to NumPy array for dataset '{dataset_name}': {e}")

                    case str() | bytes():
                        # Handle single strings or bytes
                        dt = h5py.string_dtype(encoding='utf-8')
                        f.create_dataset(dataset_name, data=np.array([data], dtype=dt))

                    case dict():
                        # Recursively save nested dictionaries
                        group = f.create_group(dataset_name)
                        save_to_hdf5_helper(data, group)
                    case _:
                        raise TypeError(f"Unsupported data type for dataset '{dataset_name}': {type(data)}")

        save_to_hdf5_helper(dataset_dict, f)

def load_hdf5(file_path: str) -> dict[str, any]:
    """
    Load data from an HDF5 file into a dictionary with proper type handling.

    :param file_path: Path to the HDF5 file.
    :return: Dictionary with dataset names as keys and data with proper types.
    """
    with h5py.File(file_path, 'r') as file:
        def load_hdf5_helper(file: h5py.File) -> dict[str, any]:
            data = {}
            for key, val in file.items():
                if isinstance(val, h5py.Group):
                    # Recursively load groups as nested dictionaries
                    data[key] = load_hdf5_helper(val)
                elif isinstance(val, h5py.Dataset):
                    # Check the dtype of the dataset
                    if val.dtype.kind in {'U', 'S'}:  # String or bytes
                        data[key] = val.asstr()[...]
                    elif val.shape == ():  # Scalar
                        data[key] = val[()]
                    else:
                        data[key] = val[...]
            # Add attributes as additional keys
            for attr_key, attr_val in file.attrs.items():
                data[attr_key] = attr_val
            return data

        return load_hdf5_helper(file)


def mean_below_diagonal(matrix: np.ndarray) -> float:
    """
    Calculate the mean of elements below the diagonal of a symmetric matrix.

    :param matrix: Symmetric numpy array with 1s on the diagonal.
    :return: Mean of the elements below the diagonal.
    """
    below_diag_elements = matrix[np.tril_indices_from(matrix, k=-1)]
    mean_value = below_diag_elements.mean()
    return float(mean_value)


def soft_dice_score(output: torch.Tensor,
                    target: torch.Tensor,
                    smooth: float = 0.0,
                    eps: float = 1e-7,
                    dims=None) -> torch.Tensor:
    """
    Compute the Soft Dice Score for the given output and target.

    :param output: Model output. Shape: (N, C, HxW)
    :param target: Target mask. Shape: (N, C, HxW)
    :param smooth: label smoothing value
    :param eps: epsilon value to avoid division by zero
    :param dims: dimensions to reduce. Default is None

    :return: soft dice score
    """
    assert output.size() == target.size()
    if dims is not None:
        intersection = torch.sum(output * target, dim=dims)
        cardinality = torch.sum(output + target, dim=dims)
    else:
        intersection = torch.sum(output * target)
        cardinality = torch.sum(output + target)
    dice_score = (2.0 * intersection + smooth) / (cardinality + smooth).clamp_min(eps)
    return dice_score

@check_is_image(kwarg_positions=(0, 1))
def multiclass_iou(*, pred_mask: torch.Tensor=None, # Enforce keyword arguments because one can swap pred_mask and true_mask mistakenly
                   true_mask: torch.Tensor=None,
                   ignore_channels: list = None) -> torch.Tensor:
    """
    Compute Intersection over Union (IoU) for the predicted mask and the true mask.

    :param pred_mask: predicted mask with shape (cls, H, W)
    :param true_mask: true mask with shape (cls, H, W)
    :param ignore_channels: list of channels to ignore (pass 0 to ignore background)

    :return: IoU score

    :raises ValueError: If the predicted mask is not a 3D tensor
    :raises ValueError: If the shape of the predicted mask and the true mask is not the same
    """
    if not len(pred_mask.shape) == len(true_mask.shape) == 3:
        raise ValueError("The predicted mask and the true mask should be 3D tensors with the same shape."
                         f"Got shape: {pred_mask.shape} for prediction and {true_mask.shape} for ground truth.")

    return IoU(ignore_channels=ignore_channels)(pred_mask, true_mask)


def get_enum_member(cls_of_interest: str, enum_class: Type[Enum]) -> Optional[Enum]:
    """
    Retrieve an enum member by its name (case-insensitive) from a given enum class.

    :param cls_of_interest: The name of the enum member to look up.
    :param enum_class: The enum class to search within.

    :returns: The corresponding enum member if found, otherwise None.
    """
    cls_name = cls_of_interest.upper()
    return enum_class.__members__.get(cls_name)

def cosine_similarity(x: np.ndarray, y: np.ndarray) -> np.ndarray:
    """
    Compute the cosine similarity between two matrices.

    :param x: First matrix
    :param y: Second matrix

    :return: Cosine similarity matrix
    """
    if isinstance(x, torch.Tensor):
        x = x.cpu().numpy()
    if isinstance(y, torch.Tensor):
        y = y.cpu().numpy()
    x = x.reshape(1, -1) if len(x.shape) == 1 else x
    y = y.reshape(1, -1) if len(y.shape) == 1 else y
    if x.shape[-1] <= 1 or y.shape[-1] <= 1:
        raise ValueError(f"Cosine similarity requires at least 2 features. Got {x.shape[-1]} features for x and {y.shape[-1]} features for y.")

    return cs(x, y)


def cluster_images_and_generate_statistics(
    features: np.ndarray,
    true_labels: np.ndarray,
    n_clusters: int,
    method: str = 'kmeans',
    **kwargs
) -> dict[str, float]:
    """
    Clusters the given features and computes ARI, NMI.

    :param features: (N, D) array of feature vectors
    :param true_labels: (N,) array of ground truth class labels
    :param n_clusters: Number of clusters to find
    :param method: 'kmeans', 'dbscan', or 'spectral'
    :param kwargs: Additional parameters for the clustering method
    :return: Dictionary of statistics {'ari': ..., 'nmi': ...}
    """
    cluster_labels = cluster_and_return_labels(
        data=features,
        method=method,
        n_clusters=n_clusters if method != 'dbscan' else None,
        **kwargs
    )

    return {
        "ri": rand_score(true_labels, cluster_labels),
        "ari": adjusted_rand_score(true_labels, cluster_labels),
        "nmi": adjusted_mutual_info_score(true_labels, cluster_labels)
    }


def plot_and_save_heatmap(matrix: Union[list, np.ndarray, torch.Tensor],
                          figsize: tuple[int, int]=None,
                          x_tick_labels: list[str]=None,
                          y_tick_labels: list[str]=None,
                          cbar_kws: dict[str, str]=None,
                          title: str="Heatmap",
                          x_label: str="X Axis",
                          y_label: str="Y Axis",
                          show: bool=True,
                          save_fig_path: str=None) -> None:
    """
    Plot a heatmap using the specified matrix.

    :param matrix: matrix
    :param figsize: figure size
    :param x_tick_labels: x-axis tick labels
    :param y_tick_labels: y-axis tick labels
    :param cbar_kws: colorbar keyword arguments
    :param title: title of the plot
    :param x_label: x-axis label
    :param y_label: y-axis label
    :param show: whether to display the plot
    :param save_fig_path: Path to save the figure
    **kwargs: Additional keyword arguments (currently available: title, xlabel, ylabel)
    """
    figsize = (len(matrix) * 0.7, len(matrix) * 0.7) if figsize is None else figsize
    plt.figure(figsize=figsize)
    sns.heatmap(matrix, annot=True, fmt=".2f", cmap="viridis",
                xticklabels=x_tick_labels if x_tick_labels else list(range(matrix.shape[1])),
                yticklabels=y_tick_labels if y_tick_labels else list(range(matrix.shape[0])),
                cbar_kws=cbar_kws if cbar_kws else {"label": "value"})
    plt.title(title)
    plt.xlabel(x_label)
    plt.ylabel(y_label)
    if save_fig_path:
        plt.savefig(save_fig_path)
    if show:
        plt.show()
    plt.close()

def plot_and_save_barplot(data: dict[str, list[float]],
                            bar_labels: list[str],
                            title: str = "Barplot",
                            xlabel: str = "X-axis",
                            ylabel: str = "Y-axis",
                            save_path: str = None,
                            show: bool = True) -> None:
    """
    Plot and save a barplot.
    :param data: Dictionary containing data to plot.
    :param bar_labels: Labels that will be displayed in the legend.
    :param title: Title of the plot.
    :param xlabel: Label for the x-axis.
    :param ylabel: Label for the y-axis.
    :param save_path: Path to save the plot image. If None, plot is not saved.
    :param show: Whether to display the plot.
    """
    x_labels = list(data.keys())
    values = list(data.values())
    num_groups = len(values[0])

    if not all(len(v) == num_groups for v in values):
        raise ValueError("All lists in data must have the same length as the number of bar labels.")

    x = np.arange(len(x_labels))  # the label locations
    width = 0.8 / num_groups      # width of each bar

    plt.figure(figsize=(10, 6))

    for i in range(num_groups):
        heights = [v[i] for v in values]
        plt.bar(x + i * width, heights, width, label=bar_labels[i])

    plt.title(title)
    plt.xlabel(xlabel)
    plt.ylabel(ylabel)
    plt.xticks(x + width * (num_groups - 1) / 2, x_labels)  # Center the tick labels
    plt.legend()
    plt.grid(axis='y', linestyle='--', alpha=0.6)

    if save_path:
        plt.savefig(save_path)

    if show:
        plt.show()

    plt.close()


def plot_and_save_lineplot(y: np.ndarray,
                           x: np.ndarray = None,
                           y_lim: tuple = None,
                           x_lim: tuple = None,
                           save_path: str = None,
                           sort_y: bool = False,
                           title: str = "Lineplot",
                           xlabel: str = "x-axis",
                           ylabel: str = "y-axis") -> None:
    """
    Plot and save a lineplot, limiting x-ticks to at most 20 evenly distributed values.

    :param y: Array of y-values.
    :param x: Array of x-values. If None, indices of y will be used.
    :param y_lim: Tuple for y-axis limits (min, max). Optional.
    :param x_lim: Tuple for x-axis limits (min, max). Optional.
    :param save_path: Path to save the plot image. If None, plot is not saved.
    :param sort_y: Whether to sort y-values before plotting.
    :param title: Title of the plot.
    :param xlabel: Label for the x-axis.
    :param ylabel: Label for the y-axis.
    :returns: None
    """
    if x is None:
        x = np.arange(len(y))

    if sort_y:
        y = np.sort(y)

    plt.figure(figsize=(10, 6))
    plt.plot(x, y, marker='o', linestyle='-', color='b')
    plt.title(title)
    plt.xlabel(xlabel)
    plt.ylabel(ylabel)
    plt.grid(True)

    if len(x) > 20:
        tick_indices = np.linspace(0, len(x) - 1, 20, dtype=int)
        tick_labels = [x[i] for i in tick_indices]
        plt.xticks(tick_indices, tick_labels, rotation=90)
    if y_lim:
        plt.ylim(y_lim)
    if x_lim:
        plt.xlim(x_lim)
    plt.tight_layout()
    if save_path:
        plt.savefig(save_path)
    plt.show()
    plt.close()


def plot_and_save_histogram(data: np.ndarray,
                            num_bins: int = 10,
                            title: str = "Histogram",
                            x_label: str = "Value",
                            y_label: str = "Frequency",
                            save_path: str = None,
                            x_lim: tuple = (0, 1),
                            show: bool = True) -> None:
    """
    Plot and save a histogram, allowing the user to choose the number of bins.

    :param data: Array of data to plot.
    :param num_bins: Number of bins for the histogram.
    :param title: Title of the plot.
    :param x_label: Label for the x-axis.
    :param y_label: Label for the y-axis.
    :param save_path: Path to save the plot image. If None, plot is not saved.
    :param x_lim: Tuple for x-axis limits (min, max). Optional.
    :param show: Whether to display the plot.
    """
    plt.figure(figsize=(10, 6))
    plt.hist(data, bins=num_bins, color='blue', edgecolor='black', alpha=0.7)
    plt.title(title)
    plt.xlabel(x_label)
    plt.ylabel(y_label)
    plt.grid(axis='y', linestyle='--', alpha=0.6)
    if x_lim:
        plt.xlim(*x_lim)
    if save_path:
        plt.savefig(save_path)
    if show:
        plt.show()
    plt.close()


def fit_regression_line(x: np.ndarray, y: np.ndarray, poly_degree: int) -> object:
    """
    Fit a polynomial regression line to the given data.

    :param x: Independent variable values (numpy array).
    :param y: Dependent variable values (numpy array).
    :param poly_degree: Degree of the polynomial regression.
    :return: RegressionResult containing predictions, coefficients, and intercept. Attributes: predictions, coefficients, intercept.
    """
    @dataclass
    class RegressionResult:
        predictions: np.ndarray
        coefficients: np.ndarray
        intercept: float
        mse: float

    poly_features = np.vander(x, N=poly_degree + 1, increasing=True)
    reg = LinearRegression().fit(poly_features, y)
    predictions = reg.predict(poly_features)
    mse = mean_squared_error(y, predictions)
    return RegressionResult(predictions, reg.coef_, reg.intercept_, mse)


def get_statistics(x: np.ndarray, y: np.ndarray) -> object:
    """
    Calculate statistics such as Pearson and Spearman correlations, standard deviation,
    mean, median, and number of data points for the given data.

    :param x: Independent variable values (numpy array).
    :param y: Dependent variable values (numpy array).
    :return: Statistics containing computed statistical encoders. Attributes: pearson, spearman, std, mean, median, n_points.
    """
    @dataclass
    class Statistics:
        pearson: float
        spearman: float
        std: float
        mean: float
        median: float
        n_points: int

    pearson, _ = pearsonr(x, y)
    spearman, _ = spearmanr(x, y)
    std, mean, median, n_points = np.std(y), np.mean(y), np.median(y), len(y)
    return Statistics(pearson, spearman, std, mean, median, n_points)


def plot_boxplot_with_regression(x: np.ndarray,
                                 y: np.ndarray,
                                 poly_degree: int = 1,
                                 x_lim: tuple = (0, 1),
                                 y_lim: tuple = (0, 1),
                                 num_bins: int = 20,
                                 title: str = "Boxplot with Regression",
                                 x_label: str = "IoU Difference",
                                 y_label: str = "Similarity Score",
                                 save_fig_path: str = None,
                                 plot_bin_regression: bool = False,
                                 verbose: bool = False,
                                 return_results: bool = False):
    """
    Plot a boxplot of y values binned by x, and add a polynomial regression line over the scatter.

    :param x: Array of x values.
    :param y: Array of y values.
    :param poly_degree: The degree of the polynomial for regression.
    :param x_lim: Tuple of (min, max) x-axis limits.
    :param y_lim: Tuple of (min, max) y-axis limits.
    :param num_bins: Number of bins for the boxplot.
    :param title: Plot title.
    :param x_label: Label for the x-axis.
    :param y_label: Label for the y-axis.
    :param save_fig_path: Optional path to save the plot.
    :param plot_bin_regression: If True, plots regression lines within each bin and prints coefficients.
    :param verbose: If True, prints additional information.
    :param return_results: If True, returns the overall statistics and per-bin statistics.
    """

    def bin_data(x: np.ndarray, y: np.ndarray, lower: float, upper: float, num_bins: int):
        bins = np.linspace(lower, upper, num_bins + 1)
        bin_indices = np.digitize(x, bins) - 1
        binned_y = [[] for _ in range(num_bins)]
        bin_centers = 0.5 * (bins[:-1] + bins[1:])
        for xi, yi, bi in zip(x, y, bin_indices):
            if 0 <= bi < num_bins:
                binned_y[bi].append(yi)
        return bin_centers, binned_y

    lower, upper = x_lim
    bin_centers, binned_y = bin_data(x, y, lower, upper, num_bins)

    # Replace empty bins with NaN to stretch the boxplots evenly
    binned_y = [b if b else [np.nan] for b in binned_y]

    plt.figure(figsize=(12, 8))
    plt.boxplot(binned_y, positions=bin_centers, widths=(upper - lower) / (num_bins * 2), patch_artist=True)

    # Apply regression if valid data exists
    valid = ~np.isnan(x) & ~np.isnan(y)
    x_valid = x[valid]
    y_valid = y[valid]

    if not len(x_valid) > 1:  # Ensure enough data points are available
        raise ValueError("Less than two data points are valid. Data is invalid for plotting.")

    regression_result = fit_regression_line(x_valid, y_valid, poly_degree)

    # Regression line
    x_line = np.linspace(lower, upper, 100)
    y_line = np.polyval(regression_result.coefficients[::-1], x_line) + regression_result.intercept
    plt.plot(x_line, y_line, color='red', linewidth=2, label=f'Regression line (Degree {poly_degree})')

    # Overall statistics
    overall_stats = get_statistics(x_valid, y_valid)

    # Display stats
    plt.text(0.05, 0.95, f"Pearson Correlation: {overall_stats.pearson:.2f}", transform=plt.gca().transAxes,
             fontsize=12, verticalalignment='top', bbox=dict(boxstyle="round", alpha=0.5))
    plt.text(0.05, 0.90, f"MSE: {regression_result.mse:.4f}", transform=plt.gca().transAxes, fontsize=12,
             verticalalignment='top', bbox=dict(boxstyle="round", alpha=0.5))
    if poly_degree == 1:
        plt.text(0.05, 0.85, f"Regression Coefficients: {regression_result.coefficients[1]:.2f}",
                 transform=plt.gca().transAxes, fontsize=12, verticalalignment='top',
                 bbox=dict(boxstyle="round", alpha=0.5))

    # Per-bin regression and statistics
    per_bin_stats = []
    if plot_bin_regression or return_results:
        for i, bin_y in enumerate(binned_y):
            bin_mask = (x_valid > bin_centers[i] - (upper - lower) / (2 * num_bins)) & (
                    x_valid <= bin_centers[i] + (upper - lower) / (2 * num_bins))
            bin_x = x_valid[bin_mask]
            bin_y = y_valid[bin_mask]

            if len(bin_x) > 1:  # Perform regression only if bin has enough data
                bin_regression_result = fit_regression_line(bin_x, bin_y, poly_degree=1)
                bin_stats = get_statistics(bin_x, bin_y)
                per_bin_stats.append({
                    "bin_index": i + 1,
                    "bin_center": bin_centers[i],
                    "bin_stats": bin_stats,
                    "regression": bin_regression_result
                })

                if plot_bin_regression:
                    plt.plot(bin_x, bin_regression_result.predictions.reshape(-1, 1),
                             label=f"Bin {i + 1} coeff: {bin_regression_result.coefficients[1]:.2f}")

                if verbose:
                    print(f"""
                    Statistics of bin {i + 1}:
                      Pearson Correlation: {bin_stats.pearson:.2f}
                      Spearman Correlation: {bin_stats.spearman:.2f}
                      Standard Deviation: {bin_stats.std:.2f}
                      Mean: {bin_stats.mean:.2f}
                      Median: {bin_stats.median:.2f}
                      Number of Data Points: {bin_stats.n_points}
                      Regression Coefficients: {bin_regression_result.coefficients[1]:.2f}
                      MSE: {bin_regression_result.mse:.4f}
                    """)

    if verbose:
        print(f"""
        Overall Statistics:
          Pearson Correlation: {overall_stats.pearson:.2f}
          Spearman Correlation: {overall_stats.spearman:.2f}
          Standard Deviation: {overall_stats.std:.2f}
          Mean: {overall_stats.mean:.2f}
          Median: {overall_stats.median:.2f}
          Regression Coefficients: {[round(coeff, 2) for coeff in regression_result.coefficients]}
        """)

    plt.xticks(bin_centers, [round(center, 2) for center in bin_centers])
    plt.title(title)
    plt.xlabel(x_label)
    plt.xlim(lower, upper)
    plt.ylabel(y_label)
    plt.ylim(*y_lim)
    plt.legend()
    if save_fig_path:
        plt.savefig(save_fig_path)
    plt.show()

    if return_results:
        return {
            "overall_statistics": overall_stats,
            "regression_result": regression_result,
            "per_bin_statistics": per_bin_stats
        }


def plot_scatter_with_regression(x: np.ndarray,
                                 y: np.ndarray,
                                 x_lim: tuple = (0, 1),
                                 y_lim: tuple = (0, 1),
                                 title: str = "Scatterplot with Regression",
                                 x_label: str = "IoU Difference",
                                 y_label: str = "Similarity Score",
                                 save_fig_path: str = None) -> None:
    """
    Plot a scatterplot of x and y, and add a regression line over the scatter.

    :param x: Independent variable values (numpy array).
    :param y: Dependent variable values (numpy array).
    :param x_lim: Limits for the x-axis.
    :param y_lim: Limits for the y-axis.
    :param title: Title of the plot.
    :param x_label: Label for the x-axis.
    :param y_label: Label for the y-axis.
    :param save_fig_path: Path to save the figure. If None, the figure is displayed.
    :returns: None
    """
    lower, upper = x_lim
    valid = ~np.isnan(x) & ~np.isnan(y)
    x_valid = x[valid]
    y_valid = y[valid]

    plt.figure(figsize=(10, 6))
    plt.scatter(x_valid, y_valid, alpha=0.6, label="Data points")

    if len(x_valid) > 1:  # Ensure enough data points are available
        reg = LinearRegression().fit(x_valid.reshape(-1, 1), y_valid)
        coeff = reg.coef_[0]
        x_line = np.linspace(lower, upper, 100).reshape(-1, 1)
        y_line = reg.predict(x_line)
        plt.plot(x_line, y_line, color='red', linewidth=2, label=f'Regression line, Coefficient: {coeff:.2f}')
    else:
        print("Insufficient data points for regression.")

    plt.title(title)
    plt.xlabel(x_label)
    plt.xlim(*x_lim)
    plt.ylabel(y_label)
    plt.ylim(*y_lim)
    plt.legend()
    if save_fig_path:
        plt.savefig(save_fig_path)
    plt.show()

def is_subset(list1: list, list2: list) -> bool:
    """
    Check if list1 is a subset of list2.

    :param list1: First list to check (potential subset)
    :param list2: Second list (or tuple) to check against (potential superset)

    :returns: True if list1 is a subset of list2, False otherwise
    """
    if len(list1) > len(list2):
        raise ValueError("List1 must be have smaller or equal length than list2")
    return set(list1).issubset(list2)

def list_is_unique(lst: list) -> bool:
    """
    Check if all elements in a list are unique.

    :param lst: List to check

    :return: True if all elements are unique, False otherwise
    """
    if len(lst) <= 1:
        return True
    for i in range(len(lst)):
        for j in range(i + 1, len(lst)):
            if lst[i] == lst[j]:
                return False
    return True

def convert_to_integers(list_of_tuples: list[tuple[float, float]]) -> list[tuple[int, int]]:
    """
    Convert all elements in a list of tuples to integers.

    :param list_of_tuples: List of tuples with float values

    :return: List of tuples with integer values
    """
    return [(int(x), int(y)) for x, y in list_of_tuples]


def standardize_data(data: np.ndarray, axis: int=0) -> np.ndarray:
    """
    Standardize the given data using the formula: (x - mean) / std.

    :param data: Input data
    :param axis: Axis along which to standardize the data (for row-wise standardization, use axis=0. For column-wise standardization, use axis=1)

    :return: Standardized data
    """
    return (data - np.mean(data, axis=axis, keepdims=True)) / np.std(data, axis=axis, keepdims=True)


def save_model(model, file_path: str) -> None:
    """
    Save the trained model to a file.

    :param model: Model to save
    :param file_path: Path to save the trained model
    """
    with open(file_path, 'wb') as file:
        joblib.dump(model, file)


def load_model(file_path: str) -> object:
    """
    Load a pre-trained model from a file.

    :param file_path: Path from which to load the trained model

    :return: Trained model
    """
    with open(file_path, 'rb') as file:
        return joblib.load(file)


def copy_or_move_images(image_paths: list[str], directory: str, operation: str="copy") -> None:
    """
    Move or copy images to the specified directory.

    :param image_paths: List of image paths
    :param directory: Directory to move or copy the images
    :param operation: Operation to perform. Choose from ['copy', 'cut']. Default is 'copy'.
    """
    assert operation in ['copy', 'cut'], "Invalid operation. Choose from ['copy', 'cut']"

    if not os.path.exists(directory):
        os.makedirs(directory)

    for image in image_paths:
        if operation == "copy":
            shutil.copy(image, directory)
        elif operation == "cut":
            shutil.move(image, directory)

def average(matrix: np.ndarray | torch.Tensor) -> float:
    """
    Compute the average of the given matrix.

    :param matrix: Input matrix

    :return: Average value
    """
    return np.mean(matrix) if isinstance(matrix, np.ndarray) else torch.mean(matrix).item()


@check_is_image()
def gaussian_blur(image: np.ndarray | torch.Tensor, kernel_size: int=None, sigma: float=1.0) -> np.ndarray | torch.Tensor:
    """
    Apply Gaussian blurring to the given image.

    :param image: Input image
    :param kernel_size: Size of the kernel
    :param sigma: Standard deviation of the kernel

    :return: Blurred image
    """
    if not kernel_size:
        kernel_size = 2 * int(3 * sigma) + 1
    min_kernel_size = 2 * int(3 * sigma) + 1
    max_kernel_size = 2 * int(5 * sigma) + 1
    if not min_kernel_size <= kernel_size <= max_kernel_size:
        raise ValueError(f"Kernel radius must be between 2 * 3-5 times the standard deviation plus one. " 
                         f"In this case, it should be between {min_kernel_size} and {max_kernel_size} "
                         f"Got kernel size: {kernel_size}")
    if isinstance(image, np.ndarray):
        return cv2.GaussianBlur(image, (kernel_size, kernel_size), sigma)
    elif torch.is_tensor(image):
        return TF.gaussian_blur(image, kernel_size, sigma).clip(0.0, 1.0)


@check_is_image()
def plot_image(image: np.ndarray | torch.Tensor, title: str = None) -> None:
    """
    Plot the image with its file path and label.
    If image shape of (3, width, height) is passed, it is converted to (width, height, 3) before plotting.
    """
    if image.shape[0] == 3:
        image = np.transpose(image, (1, 2, 0))
    plt.imshow(image)
    plt.title(title)
    plt.axis('off')
    plt.show()


@check_is_image()
def denoise_mask(mask: np.ndarray, min_size: int) -> np.ndarray:
    """
    Denoises the input binary mask image by removing components smaller than the specified minimum size.

    :param mask: Input binary mask (HxW)
    :param min_size: Minimum pixel size for components to retain

    :returns: Processed mask with only components of size >= min_size, retaining original class values.
    """
    filtered_mask = np.zeros_like(mask, dtype=np.uint8)

    for class_value in np.unique(mask):
        logging.debug("List of classes before being filtered: %s", class_value)
        if class_value == 0:  # Skip the background class
            continue

        class_mask = np.where(mask == class_value, 255, 0).astype(
            np.uint8)  # Classes of question are set to 255. The rest is 0.
        num_labels, labels, stats, _ = cv2.connectedComponentsWithStats(class_mask, connectivity=4)
        for i in range(1, num_labels):
            component_size = stats[i, cv2.CC_STAT_AREA]
            if component_size >= min_size:
                filtered_mask[labels == i] = class_value

    logging.debug("List of classes after being filtered: %s", np.unique(filtered_mask))
    return filtered_mask




import abc
import logging

import numpy as np

class SimilarityMetric(abc.ABC):
    """
    Abstract base for all similarity encoders.

    All concrete similarity metric classes must inherit from this class.
    """
    _logger = logging.getLogger('Similarity_Metrics')
    @abc.abstractmethod
    def similarity_score(self, image1: np.ndarray, image2: np.ndarray) -> float:
        """
        Compute a similarity score between two images.

        :param image1: First image
        :param image2: Second image
        :return: A similarity score
        """
        pass

class FeatureExtractorBase(abc.ABC):
    """
    Abstract interface for extracting features from images.

    A feature extractor transforms an image (NumPy array) into a
    set of feature vectors (NumPy array).
    """
    _logger = logging.getLogger("Feature_Extractor")
    def __init__(self):
        pass

    @abc.abstractmethod
    def __call__(self, image: np.ndarray) -> np.ndarray:
        """
        Extracts features from an image.

        :param image: Input image (NumPy array).
        :return: Feature descriptors (NumPy array).
        """
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def output_dim(self) -> int:
        """
        The dimensionality (D) of each feature vector, i.e., shape[1] of the output.
        """
        raise NotImplementedError


from collections.abc import Iterator
import logging
from itertools import tee
from typing import Iterable, Callable

import numpy as np
import torch
import cv2

from ..encoders._base_encoder import ImageEncoderBase, SimilarityMetric
from .._utils import cosine_similarity
from ._base_encoder import check_desired_output


class Pipeline(SimilarityMetric):
    """
    A pipeline for computing feature vectors using a set of
    descriptor-based encoders (e.g., VLAD, Fisher, etc.).

    Currently, all vectors computed using the Encoders listed
    will always be flattened, because different Encoders also
    have different output sizes.

    :param encoders: A list of ImageEncoderBase instances.
    :param similarity_func: A function that takes two batches of vectors and returns a similarity score
    matrix with size (batch_1_size, batch_2_size).
    """
    _logger = logging.getLogger("Pipeline")
    def __init__(
            self,
            encoders: list[ImageEncoderBase],
            similarity_func: Callable[[np.ndarray, np.ndarray], float] = cosine_similarity,
    ):
        self._check_valid_encoders(encoders)
        self.encoders = encoders
        self._similarity_func = similarity_func

    def _check_valid_encoders(self, encoders: list[ImageEncoderBase]) -> None:
        """
        Checks if all encoders in the pipeline are instances of ImageEncoderBase.
        :param encoders: list of encoders to check.
        """
        for encoder in encoders:
            if not isinstance(encoder, ImageEncoderBase):
                raise ValueError(f"Pipeline only accepts instances of ImageEncoderBase, not {type(encoder)}")

    def encode(self, images: Iterable[np.ndarray] | np.ndarray) -> np.ndarray:
        """
        Encode an image using all encoders in the pipeline.

        :param images: Images to process.
        :return: encoded images using the combined encoders.
        """
        all_encodings = []
        if isinstance(images, torch.Tensor):
            raise RuntimeError("Torch images are not supported yet.")
        if isinstance(images, np.ndarray) and images.ndim == 3:
            images = [images] # Handle single image case
        images = tee(images, len(self.encoders))
        for metric in self.encoders:
            a = metric.flatten # each encoder has to be flattened to be usable here. Saving the original state temporarily
            metric.flatten = True
            encodings = metric.encode(images) # Each of size (num_imgs, feature_dim)
            all_encodings.append(encodings)
            metric.flatten = a
        return np.hstack(all_encodings)

    def generate_encoding_map(self, image_paths: Iterable[str]) -> dict[str, np.ndarray]:
        """
        Converts a collection of image file paths into a dictionary of
        ``{image_path: encoded_vector}``.

        This method automatically reads each image, applies the internal
        encoding pipeline, and stores the resulting descriptor vector.

        :param image_paths: List of image full paths
        :return: a dictionary where keys are image paths and values are descriptor vectors of the
                corresponding images
        """
        images = (cv2.cvtColor(cv2.imread(path), cv2.COLOR_BGR2RGB) for path in image_paths)
        return dict(zip(image_paths, self.encode(images)))

    @property
    def similarity_func(self):
        return self._similarity_func

    @similarity_func.setter
    def similarity_func(self, func: Callable[[np.ndarray, np.ndarray], float]):
        dummy1, dummy2 = np.random.rand(10, 10), np.random.rand(10, 10)
        self._similarity_func = check_desired_output(func, dummy1, dummy2)

    def similarity_score(self, images1: Iterable[np.ndarray] | np.ndarray, images2: Iterable[np.ndarray] | np.ndarray) -> float:
        """
        Computes vector encodings for two images and calculates the similarity score between them.

        :param images1: First (batch of) image(s)
        :param images2: Second (batch of) image(s)
        :return: Similarity score. If image iterables are provided, a similarity matrix between two image batches is returned.
        """
        vector1 = self.encode(images1)
        vector2 = self.encode(images2)
        result = self.similarity_func(vector1, vector2)
        return np.float_(result)

    # def fit(self, images: Iterable[np.ndarray], reduce_dimension: bool = False, reduce_factor: int=2) -> None:
    #     """
    #     Trains any clustering model_files used by the encoders in this pipeline, if they have a fit method.
    #
    #     :param images: Iterable of images (NumPy arrays) used for fitting the pipeline's encoders.
    #     :param reduce_dimension: Whether to apply dimension reduction (e.g., PCA) if supported.
    #     :param reduce_factor: Factor to reduce the dimension by.
    #     """
    #     for metric in self.encoder:
    #         if hasattr(metric, 'fit') and callable(metric.fit):
    #             self._logger.info(f"Fitting {metric.__class__.__name__} with reduce_dimension={reduce_dimension}...")
    #             metric.fit(images, reduce_dimension=reduce_dimension, reduce_factor=reduce_factor)
    #         else:
    #             self._logger.warning(f"{metric.__class__.__name__} has no 'fit' method. Skipping...")

    def __repr__(self) -> str:
        """
        Returns a string representation of this Pipeline, including the names
        of the encoders and the similarity function used.
        """
        encoders_str = "\n".join([str(encoder) for encoder in self.encoders])
        return (f"Pipeline(\n"
                f"encoders=[{encoders_str}],\n"
                f"similarity_func={self._similarity_func.__name__ if hasattr(self._similarity_func, '__name__') else str(self._similarity_func)})")


from setuptools import setup, find_packages

setup(
    name="pyvisim",
    version="0.1.1rc",
    author="Nhat Huy Vu",
    author_email="vunhathuy234@gmail.com",
    description="A Python library for image similarity analysis using Image Encoders and Neural Networks",
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    url="https://github.com/MechaCritter/Python-Visual-Similarity",
    license="MIT",
    packages=find_packages(include=["pyvisim", "pyvisim.*"]),
    python_requires=">=3.10",
    install_requires=[
        "opencv-python",
        "h5py",
        "numpy",
        "opencv-python",
        "pyaml",
        "PyYAML",
        "scikit-learn",
        "scipy",
        "seaborn",
        "torch",
        "torchvision",
        "torchaudio"
    ],
    classifiers=[
        "Programming Language :: Python :: 3",
        "Operating System :: OS Independent",
    ],
)
